import numpy as np

def calc_R_lengths(blockstart, blockend, rec_rate_per_chunk, calc_start, calc_end, chunk_size):
    """
    Calculates the weighted lengths of each conserved block (gene), so that for example if the mean 
    recombination rate across the block is 0.5, this will return the length of the block multiplied by 0.5
    """
    num_precise_chunks = (calc_end - calc_start - 1) // chunk_size # -1 to fix extra chunk bug. May need to revert.

    # Build chunk boundaries (note: length = num_chunks + 1)
    chunk_starts = calc_start + np.arange(0, num_precise_chunks + 1) * chunk_size
    next_chunk_starts = chunk_starts + chunk_size  # shape: (num_chunks+1,)

    # Compute the overlap between each block and each chunk interval:
    # For block i and chunk j, the overlap is:
    #   max(0, min(blockend[i], chunk_right[j]) - max(blockstart[i], chunk_left[j]))
    overlap = np.maximum(0, np.minimum(blockend[:, None], next_chunk_starts[None, :]) -
                           np.maximum(blockstart[:, None], chunk_starts[None, :]))
    weighted_overlap = overlap * rec_rate_per_chunk[None, :] # Multiply by the recombination rate for each chunk
    weighted_sum = np.sum(weighted_overlap, axis=1) # Sum over the chunk intervals for each block
    
    return weighted_sum

def calc_R_distances(
    precise_blockstart, precise_blockend, precise_rates,
    precise_region_start, precise_region_end, chunk_size,
    pos_chunk_clean, chunk_start
):
    

    num_precise_chunks = (precise_region_end - precise_region_start) // chunk_size
    chunk_starts = precise_region_start + np.arange(0, num_precise_chunks + 1) * chunk_size
    chunk_ends = np.minimum(chunk_starts + chunk_size, precise_region_end)
    this_chunk_idx = np.where(chunk_starts == chunk_start)[0][0]
    chunk_end = chunk_ends[this_chunk_idx]

    blockstart_chunks = (precise_blockstart - precise_region_start) // chunk_size
    blockend_chunks = (precise_blockend - precise_region_start) // chunk_size

    pos_broadcast = pos_chunk_clean[None, :]  # shape (1, P)
    # BLOCKEND (upstream) distances
    inchunk_distances_end = np.minimum(
        pos_broadcast - precise_blockend[:, None],
        pos_broadcast - chunk_start
    )
    rec_distance_focal_end = inchunk_distances_end * precise_rates[this_chunk_idx]

    # Mask out focal distances if block is in a future chunk
    same_or_earlier = (blockend_chunks <= this_chunk_idx)[:, None]
    rec_distance_focal_end *= same_or_earlier

    is_diffchunk_end = (blockend_chunks < this_chunk_idx).astype(float)[:, None]
    chunk_edge_dist_end = chunk_ends[blockend_chunks] - precise_blockend
    rec_blockchunk_end = chunk_edge_dist_end * precise_rates[blockend_chunks]
    blockend_overlap_dists = np.array([
        np.sum(precise_rates[blockend_chunks[i]+1:this_chunk_idx] * chunk_size)
        for i in range(len(blockend_chunks))
    ])
    blockend_rec_distances = rec_distance_focal_end + is_diffchunk_end * (
        rec_blockchunk_end[:, None] + blockend_overlap_dists[:, None]
    )

    # BLOCKSTART (downstream) distances
    inchunk_distances_start = np.minimum(
        precise_blockstart[:, None] - pos_broadcast,
        chunk_end - pos_broadcast
    )
    rec_distance_focal_start = inchunk_distances_start * precise_rates[this_chunk_idx]

    # Mask out focal distances if block is in a previous chunk
    same_or_later = (blockstart_chunks >= this_chunk_idx)[:, None]
    rec_distance_focal_start *= same_or_later

    is_diffchunk_start = (blockstart_chunks > this_chunk_idx).astype(float)[:, None]
    chunk_edge_dist_start = precise_blockstart - chunk_starts[blockstart_chunks]
    rec_blockchunk_start = chunk_edge_dist_start * precise_rates[blockstart_chunks]
    blockstart_overlap_dists = np.array([
        np.sum(precise_rates[this_chunk_idx+1:blockstart_chunks[i]] * chunk_size)
        for i in range(len(blockstart_chunks))
    ])
    blockstart_rec_distances = rec_distance_focal_start + is_diffchunk_start * (
        rec_blockchunk_start[:, None] + blockstart_overlap_dists[:, None]
    )

    return blockend_rec_distances, blockstart_rec_distances

def calc_R_lendist_for_chunks(upstream_indices, downstream_indices, rec_rate_per_chunk, relevant_upstream_psdc_lengths, relevant_downstream_psdc_lengths, chunk_idx, chunk_size, relevant_upstream_pseudoblockends, relevant_downstream_pseudoblockstarts, chunk_starts, chunk_ends, chunk_rec_distances, num_chunks):

    ## Calculate relevant upstream and downstream rec lengths of pseudoblocks
    upstream_rec_rates = rec_rate_per_chunk[upstream_indices] # Relevant rec rates for pseudochunks upstream
    upstream_rec_lengths = upstream_rec_rates * relevant_upstream_psdc_lengths
    downstream_rec_rates = rec_rate_per_chunk[downstream_indices] # Relevant rec rates for pseudochunks downstream
    downstream_rec_lengths = downstream_rec_rates * relevant_downstream_psdc_lengths

    ## Calculate relevant upstream rec distances!
    mean_rec_distance_focalchunk = rec_rate_per_chunk[chunk_idx] * chunk_size / 2 - 0.5 # Note that this is distance to middle of focal chunk.

    if chunk_idx == num_chunks -1: # If it's the final chunk (which may be not be full chunk_size length)
        end_focalchunk_distance = (chunk_ends[chunk_idx] - chunk_starts[chunk_idx])/2 # I think needs to be shifted by 1bp
        mean_rec_distance_focalchunk = end_focalchunk_distance

    upstream_distance_blockchunk = chunk_ends[upstream_indices] - relevant_upstream_pseudoblockends
    upstream_rec_distance_blockchunk = upstream_distance_blockchunk * rec_rate_per_chunk[upstream_indices] # This is rec distance from edge of pseudoblock to its chunk end

    chunk_cumsum = np.concatenate([[0], np.cumsum(chunk_rec_distances)])
    upstream_start = np.array(upstream_indices) + 1
    upstream_end = chunk_idx
    upstream_overlapped_rec_distances = chunk_cumsum[upstream_end] - chunk_cumsum[upstream_start]  # This is rec distance spanned in fully overlapped chunks

    upstream_rec_distances = mean_rec_distance_focalchunk + upstream_rec_distance_blockchunk + upstream_overlapped_rec_distances # Combined rec distance from middle of focal chunk to edge of pseudo"blocks" upstream

    ## Calculate downstream rec distances!
    downstream_distance_blockchunk = relevant_downstream_pseudoblockstarts - chunk_starts[downstream_indices]
    downstream_rec_distance_blockchunk = downstream_distance_blockchunk * rec_rate_per_chunk[downstream_indices] # This is rec distance from edge of pseudoblock to its chunk start
    
    downstream_start = chunk_idx + 1
    downstream_end = np.array(downstream_indices)
    downstream_overlapped_rec_distances = chunk_cumsum[downstream_end] - chunk_cumsum[downstream_start] # This is rec distance spanned in fully overlapped chunks

    downstream_rec_distances = mean_rec_distance_focalchunk + downstream_rec_distance_blockchunk + downstream_overlapped_rec_distances # Combined rec distance from middle of focal chunk to edge of pseudo"blocks" upstream

    return upstream_rec_lengths, downstream_rec_lengths, upstream_rec_distances, downstream_rec_distances


#This script is to get the slope of recovery of pi and the number of  base pairs that would lead to 50%, 75% and 90% recovery for a given recombination rate.
#I'm here multiplying both the recombination and gene conversion rate by 0.5, just to see more of an effect of demography.

import sys
import math
import numpy as np

#Define constants needed from the user:
g = 0.5*1e-8 #1e-8 #rate of gene conversion
tract_len=440 #mean tract length of gene conversion in base pairs
r = 0.5*1.0*1e-8 #rate of recombination
l = 10000.0 #(*Length of genomic element*)
u = 3.0*1e-9 #(*Mutation rate*)
#Parameters of instantaneous change in demographic history:
Nanc = 1e6 #(Ancestral population size)
Ncur = 1e6 #(Current population size)
TIME="T_0_5" #T_0_1/T_0_5/T_1
time_of_change=0.0 #0.1/0.5/1(This is the time of change in 2Ncur generations in the past.)
#Parameters of the DFE:
DFE="DFE1"
f0 = 0.1 #(*Proportion of effectively neutral mutations with 0 <= |2Nes| < 1 *)
f1 = 0.7 #(*Proportion of weakly deleterious mutations with 1 <= |2Nes| < 10 *)
f2 = 0.1 #(*Proportion of moderately deleterious mutations with 10 <= |2Nes| < 100 *)
f3 = 0.1 #(*Proportion of strongly deleterious mutations with |2Nes| >= 100 *)
#(*Note that the number of classes can easily be increased to whatever is required to approximate the continuous DFE *)
h = 0.5 #(* dominance coefficient, default can be 0.5 *)
s_window_size = 100

#Constants that we do not need from the user:
U = l*u
gamma_cutoff = 2.0
pi = 4*Nanc*u #(*Expected nucleotide diversity under neutrality*)
#(*Now we define the boundaries of the fixed intervals over which we will integrate *)
t0 = 0.0
t1 = h*(1/(2*Nanc))
t1half = h*(gamma_cutoff/(2*Nanc)) #(* This is the cut-off value of 2Nes=5. This derivation assumes that all mutations with 2Nes<5 will not contribute to BGS *)
t2 = h*(10/(2*Nanc))
t3 = h*(100/(2*Nanc))
t4 = h*1.0

def calculate_a_and_b(posn):
    C = (1.0 - math.exp(-2.0*r*posn))/2.0
    if g==0:
        a = C
        b = C + r*l
    elif g > 0:
        if posn+l < 0.5*tract_len:#The 0.5 is currently aribtrary. It's just about when the approximation of 1-exp(-x)~x holds. 
            #print ("accounting for small-distance gene conversion")
            a = (C + (g*posn))
            b = (C + r*l + (g*(posn+l)))
        else:
            #print ("accounting for large-distance gene conversion")
            a = g*tract_len + C
            b = g*tract_len + r*l + C
    return (a, b)

def calculate_exponent(t_start, t_end, posn):
    t_tmp = calculate_a_and_b(posn)
    a = t_tmp[0]
    b = t_tmp[1]
    E1 = ((U*a)/((1-a)*(a-b)*(t_end-t_start))) * math.log((a+(t_end*(1-a)))/(a + (t_start*(1-a))))
    E2 = -1.0*((U*b)/((1-b)*(a-b)*(t_end-t_start)))*math.log((b + ((1-b)*t_end))/(b + ((1-b)*t_start)))
    E = E1 + E2
    return (E)

def calculate_B(posn):
    E_f1 = calculate_exponent(t1half, t2, posn) 
    E_f2 = calculate_exponent(t2, t3, posn)
    E_f3 = calculate_exponent(t3, t4, posn)
    E_bar = f0*0.0 + f1*((t1half-t1)/(t2-t1))*0.0 + f1*((t2-t1half)/(t2-t1))*E_f1 + f2*E_f2 + f3*E_f3
    B = math.exp(-1.0*E_bar)
    return (B)


def calculate_pi(posn):
    pi_posn = pi*calculate_B(posn)
    return (pi_posn)

def calculate_pi_window(win_start, win_end):
    i=win_start
    pi_sum = 0.0
    while i <= win_end:
        pi_sum = pi_sum + float(calculate_pi(i))
        i = i + 1
    return(pi_sum/(win_end - win_start+1))

def calculate_Banc_window(win_start, win_end):
    i=win_start
    B_sum = 0.0
    while i <= win_end:
        B_sum = B_sum + float(calculate_B(i))
        i = i + 1
    return(B_sum/(win_end - win_start+1))

def get_Bcur(Banc):
    R = float(Nanc)/float(Ncur)
    Bcur = (Banc*(1.0 + (R-1.0)*math.exp((-1.0*time_of_change)/Banc))) / (1 + (R-1.0)*math.exp(-1.0*time_of_change))
    return (Bcur)

                                                                                
#Getting an average pi over a window:
#Output would need to be 3 columns: l_start, l_end, and l_B
l_start, l_end, l_B = [], [], [] #lists
posn_start = 1
len_intergenic = 10000 #fixed constant that can be changed later
posn_start = 1
while posn_start <= len_intergenic:
    l_start.append(posn_start)
    l_end.append(posn_start+s_window_size-1)
    l_B.append(get_Bcur(calculate_Banc_window(posn_start, posn_start+s_window_size)))
    posn_start = posn_start + s_window_size

print (l_start)
print(l_end)
print(l_B)
print("done")





import os
import numpy as np
import importlib.util

def _load_params_module():
    """
    Dynamically load the population parameters from BCALC_params.
    """
    params_path = os.environ.get("BCALC_params")
    if not params_path:
        raise KeyError(
            "Environment variable BCALC_params not set. Cannot load population parameters."
        )
    spec = importlib.util.spec_from_file_location("params", params_path)
    params = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(params)
    return params

def get_Bcur(Banc: np.ndarray) -> np.ndarray:
    """
    Apply demographic change to an array of B values using parameters from the population parameters module.

    Args:
        Banc: np.ndarray of B under ancestral population size

    Returns:
        np.ndarray of B under current population size
    """
    pop = _load_params_module()
    Nanc = pop.Nanc
    Ncur = pop.Ncur
    time_of_change = pop.time_of_change

    # Ratio of ancestral to current population sizes
    R = Nanc / Ncur

    # Compute exponential terms
    exp_term_num = np.exp(-(time_of_change / (2 * Ncur)) / Banc)
    exp_term_den = np.exp(-(time_of_change / (2 * Ncur)))

    # Numerator and denominator of the demographic adjustment formula
    numerator = Banc * (1 + (R - 1) * exp_term_num)
    denominator = 1 + (R - 1) * exp_term_den

    return numerator / denominator



import sys
import argparse

def parseSiteArgs(argv=None):
    parser = argparse.ArgumentParser(description="Calculates B for a single neutral site given a distance from a single selected region and prints to console.")
    parser.add_argument('--params', type=str, required=True, help="Path to Python file with population genetic parameters, e.g., ExampleParams.py")
    parser.add_argument('--element_size', type=int, default=10000, help="Length of single region (e.g. gene) under selection. [5000]")
    parser.add_argument('--distance', type=int, default=1, help="Length of single region (e.g. gene) under selection. [5000]")
    parser.add_argument('--pop_change', action='store_true', help="If set, B will reflect the current B after a step change in population size, rather than ancestral B.")
    parser.add_argument('--gamma_dfe', action='store_true', help="If set, gamma distribution parameters will be used to define DFE's discretized f0-f3 proportions")
    parser.add_argument('--constant_dfe', action='store_true', help="If set, the constant `s` and `neu_prop` parameters in the params file will be used for all conserved regions instead of discretized f0-f3 proportions")   
    parser.add_argument('--quiet', action='store_true', help="If set, silence print statements.")
    return parser.parse_args(argv)

def parseGeneArgs(argv=None):
    parser = argparse.ArgumentParser(description="Calculates B for neutral sites flanking a single element under selection.")
    parser.add_argument('--params', type=str, required=True, help="Path to Python file with population genetic parameters, e.g., ExampleParams.py")
    parser.add_argument('--element_size', type=int, default=10000, help="Length of single region (e.g. gene) under selection. [5000]")
    parser.add_argument('--flank_len', type=int, default=40000, help="Length of flanking neutral region for which to calcuate recovery of B. [25000]")
    parser.add_argument('--pop_change', action='store_true', help="If set, B will reflect the current B after a step change in population size, rather than ancestral B.")
    parser.add_argument('--gamma_dfe', action='store_true', help="If set, gamma distribution parameters will be used to define DFE's discretized f0-f3 proportions")
    parser.add_argument('--constant_dfe', action='store_true', help="If set, the constant `s` and `neu_prop` parameters in the params file will be used for all conserved regions instead of discretized f0-f3 proportions")   
    parser.add_argument('--plot', nargs='?', const='Bplot.png', default=None, 
                        help="Generate a B recovery slope output"
                            "Provide path to plot output.")
    parser.add_argument('--out', type=str, default=None,
                        help="Optional path to output CSV file.")
    parser.add_argument('--out_binsize', type=int, default=None, help="Size of bins to write average B in. By default B is saved per-base")
    parser.add_argument('--quiet', action='store_true', help="If set, silence print statements.")
    raw = argv if argv is not None else sys.argv[1:]
    args = parser.parse_args(argv)

    if '--out' in raw and '--out_binsize' not in raw: # enforce: if they asked for --out, they must have also provided --out_binsize
        parser.error("argument --out_binsize is required when --out is specified")
    return args

def parseRegionArgs(argv=None):
    parser = argparse.ArgumentParser(description="Calculates B for all neutral sites across given chromosome.")
    # parser.add_argument('--params', type=int, required=True, help="Path to file providing popgen parameters specific to modelled population (empirical or simulated).")
    parser.add_argument('--params', type=str, required=True, help="Path to Python file with population genetic parameters, e.g., ExampleParams.py")
    parser.add_argument('--bedgff', type=str, required=True, help="Path to input BED or GFF3 file.")
    parser.add_argument('--chunk_size', type=int, default=None, help="Size of chunks calculated simultaneously (bp). Default: 20000, auto-adjusted for large datasets.")
    parser.add_argument('--chunk_batch_size', type=int, default=250, help="Number of chunks to process before flushing memory. [250]")
    parser.add_argument('--precise_chunks', type=int, default=3, help="Number of adjacent chunks to calculate B precisely.")
    parser.add_argument('--pop_change', action='store_true', help="If set, B will reflect the current B after a step change in population size, rather than ancestral B.")
    parser.add_argument('--prior_Bmap', type=str, default=None,
                        help="Optional input with per-site expected diversity, e.g. a B map calculated using Bvalcalc on different annotations! Usage: --prior_Bmap your_map.csv "
                             "These values will be multiplied by the caluculated B, i.e. a value of 0.9 at a given position will be returned as 0.9 * [newly calculated B]"
                             "Format should be the same as the B-map output: 'Chromosome,Start,Conserved,B'. "
                             "Note that the Conserved column is needed for accurate parsing but will not affect the analysis.")
    parser.add_argument('--rec_map', nargs='?', default=None,
                        help="Optional recombination (crossover) map input. Usage: --rec_map your.map, "
                             "Format should be a two column csv with the header: 'start,rate'. "
                             "Note that recombination rates will be averaged within each chunk.")    
    parser.add_argument('--gc_map', nargs='?', default=None,
                        help="Optional gene conversion (non-crossover) map input. Usage: --gc_map your.map, "
                             "Format should be a two column csv with the header: 'start,rate'. "
                             "Note that gene conversion rates will be averaged within each chunk.")    
    parser.add_argument('--gamma_dfe', action='store_true', help="If set, gamma distribution parameters will be used to define DFE's discretized f0-f3 proportions")
    parser.add_argument('--constant_dfe', action='store_true', help="If set, the constant `s` and `neu_prop` parameters in the params file will be used for all conserved regions instead of discretized f0-f3 proportions")   
    parser.add_argument('--plot', nargs='?', const='genome_plot.png', default=None, 
                        help="Generate a basic plot using `Bvalcalc --genome` output"
                            "Provide path to plot output.")
    parser.add_argument('--hri', action='store_true', help="If set, will enable post-hoc calculation of B under HRI (B'; Becher and Charlesworth 2025), for low recombination regions")   
    parser.add_argument('--neutral_only', action='store_true', help="If set, plot will only show neutral sites.")
    parser.add_argument('--out', type=str, default=None,
                        help="Required path to output CSV file. If --out is specified but no file name is given, "
                             "'b_values.csv' will be used in the current directory. If --out is not specified, "
                             "no CSV will be saved. Note that by default it is per-base B, to output B averaged across"
                             "bins, use --out_bins [int]")
    parser.add_argument('--out_binsize', type=int, default=None, help="Size of bins to write average B in. By default B is saved per-base")
    parser.add_argument('--verbose', action='store_true', help="If set, will give per-chunk summaries")
    parser.add_argument('--quiet', action='store_true', help="If set, silence print statements.")
    
    raw = argv if argv is not None else sys.argv[1:]
    args = parser.parse_args(argv)

    if '--out' in raw and '--out_binsize' not in raw: # enforce: if they asked for --out, they must have also provided --out_binsize
        parser.error("argument --out_binsize is required when --out is specified")

        
    return args

def parseGenomeArgs(argv=None):
    parser = argparse.ArgumentParser(description="Calculates B for all neutral sites across given chromosome.")
    # parser.add_argument('--params', type=int, required=True, help="Path to file providing popgen parameters specific to modelled population (empirical or simulated).")
    parser.add_argument('--params', type=str, required=True, help="Path to Python file with population genetic parameters, e.g., ExampleParams.py")
    parser.add_argument('--bedgff', type=str, required=True, help="Path to input BED or GFF3 file.")
    parser.add_argument('--chr_sizes', type=str, default=None, help="Chromosome sizes file. Defaults to end of last gene in each chromosome if not provided.")
    parser.add_argument('--chunk_size', type=int, default=None, help="Size of chunks calculated simultaneously (bp). Default: 20000, auto-adjusted for large datasets.")
    parser.add_argument('--chunk_batch_size', type=int, default=250, help="Number of chunks to process before flushing memory. [250]")
    parser.add_argument('--precise_chunks', type=int, default=3, help="Number of adjacent chunks to calculate B precisely.")
    parser.add_argument('--pop_change', action='store_true', help="If set, B will reflect the current B after a step change in population size, rather than ancestral B.")
    parser.add_argument('--prior_Bmap', type=str, default=None,
                        help="Optional input with per-site expected diversity, e.g. a B map calculated using Bvalcalc on different annotations! Usage: --prior_Bmap your_map.csv "
                             "These values will be multiplied by the caluculated B, i.e. a value of 0.9 at a given position will be returned as 0.9 * [newly calculated B]"
                             "Format should be the same as the B-map output: 'Chromosome,Start,Conserved,B'. "
                             "Note that the Conserved column is needed for accurate parsing but will not affect the analysis.")
    parser.add_argument('--rec_map', type=str, default=None,
                        help="Optional recombination (crossover) map input. Usage: --rec_map your.map, "
                             "Format should be a two column csv with the header: 'start,rate'. "
                             "Note that recombination rates will be averaged within each chunk.")    
    parser.add_argument('--gc_map', type=str, default=None,
                        help="Optional gene conversion (non-crossover) map input. Usage: --gc_map your.map, "
                             "Format should be a two column csv with the header: 'start,rate'. "
                             "Note that gene conversion rates will be averaged within each chunk.") 
    parser.add_argument('--gamma_dfe', action='store_true', help="If set, gamma distribution parameters will be used to define DFE's discretized f0-f3 proportions")   
    parser.add_argument('--constant_dfe', action='store_true', help="If set, the constant `s` and `neu_prop` parameters in the params file will be used for all conserved regions instead of discretized f0-f3 proportions")   
    parser.add_argument('--hri', action='store_true', help="If set, will enable post-hoc calculation of B under HRI (B'; Becher and Charlesworth 2025), for low recombination regions")   
    parser.add_argument('--neutral_only', action='store_true', help="If set, plot will only show neutral sites.")
    parser.add_argument('--out', type=str, default=None,
                        help="Required path to output CSV file. If --out is specified but no file name is given, "
                             "'b_values.csv' will be used in the current directory. If --out is not specified, "
                             "no CSV will be saved.")
    parser.add_argument('--out_binsize', type=int, default=None, help="Size of bins to write average B in. By default B is saved per-base")
    parser.add_argument('--verbose', action='store_true', help="If set, will give per-chunk summaries")
    parser.add_argument('--quiet', action='store_true', help="If set, silence print statements.")

    raw = argv if argv is not None else sys.argv[1:]
    args = parser.parse_args(argv)

    if '--out' in raw and '--out_binsize' not in raw: # enforce: if they asked for --out, they must have also provided --out_binsize
        parser.error("argument --out_binsize is required when --out is specified")
        
    return args

def parseBmapArgs(argv=None):
    parser = argparse.ArgumentParser(description="B-map utilities for getting B statistics for sites in a VCF/txt file with specific positions.")
    parser.add_argument('--positions', type=str, required=True, help="VCF or  input as following argument")
    parser.add_argument('--plot_distribution', nargs='?', const='B_distribution.png', default=None, help="Output path for a plot of the distribution of B across each chromosome.")   
    parser.add_argument('--out', type=str, default=None,
                        help="Path to save per-site B for variant sites in the VCF/txt file. Results are not saved if --out is not specified.")
    parser.add_argument('--out_minimum', type=float, default=None, help="If set, only the sites ABOVE the given threshold of B will be returned, i.e. B > [threshold].")   
    parser.add_argument('--out_maximum', type=float, default=None, help="If set, only the sites BELOW the given threshold of B will be returned, i.e. B < [threshold].")   
    parser.add_argument('--bcftools_format', action='store_true', help="If set, removes the B value column and reformats the output for filter with bcftools view.")
    parser.add_argument('--quiet', action='store_true', help="If set, silence print statements.")
    return parser.parse_args(argv)

def parse_args(version):
    parser = argparse.ArgumentParser(
        prog="Bvalcalc",
        description=f"Welcome to Bvalcalc v{version}! Please specify a mode to calculate B."
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--generate_params', metavar='SPECIES', nargs='?', const='template', default=None,
                       help="Save population parameters from a species template")
    group.add_argument('--site', '-s', action='store_true', help="Calculate B values for a single site from a selected element")
    group.add_argument('--gene', '-g', action='store_true', help="Calculate B values for a neutral region adjacent to a single selected element")
    group.add_argument('--region', '-r', type=str, help="Calculate B values for a specific chromosomal region, considering genome-wide effects. Provide region as [CHR,START,END].")
    group.add_argument('--genome', '-w', action='store_true', help="Calculate B values genome-wide for all sites considering all selected elements")
    group.add_argument('--Bmap', '-b', type=str, help="B-map lookup for sites in a VCF/txt file")
    group.add_argument('--download_sample_data', action='store_true', help="Download sample data files to current directory (or use --dir to specify)")
    
    parser.add_argument('--dir', '-d', default='.', help="Directory to write the generated params file or download sample data (default: current directory)")

    # If no args provided, print help and exit
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    return parser



import sys
import csv
import numpy as np

def load_Bmap(file_path):
    chromosomes = []
    positions   = []
    b_values    = []

    with open(file_path, 'r') as f:
        reader = csv.reader(f)
        for row in reader:
            if not row:
                continue
            # skip comments or stray headers
            if row[0].startswith('#') or row[0] == 'Chromosome':
                continue
            if len(row) < 3:
                # malformed line, skip
                continue

            chrom = row[0]
            pos   = int(row[1])

            # Strip trailing single quote if present
            b_str = row[2].strip()
            if b_str.endswith("'"):
                b_str = b_str[:-1]
            b = float(b_str)

            chromosomes.append(chrom)
            positions.append(pos)
            b_values.append(b)

    chromosomes = np.array(chromosomes, dtype='<U20')
    positions   = np.array(positions,   dtype=np.int64)
    b_values    = np.array(b_values,    dtype=np.float64)

    return chromosomes, positions, b_values



#!/usr/bin/env python3
import os
import sys
import time
import argparse
from Bvalcalc.utils.parseArgs import parse_args, parseGenomeArgs, parseRegionArgs, parseGeneArgs, parseSiteArgs, parseBmapArgs
from Bvalcalc.core.plotB import plotB
from Bvalcalc.core.deprecated.plotB_figures import plotB_figures
from Bvalcalc.core.deprecated.plotB_figures_200kb import plotB_figures_200kb
from Bvalcalc.utils.generateParams import SPECIES, generateParams, check_generate_params_args
from Bvalcalc.core.positionsBstats import positionsBstats
from Bvalcalc.core.plotChromB import plotChromB
from Bvalcalc.utils.sample_data import download_sample_data

__version__ = "0.7.2"

def main():
    start_time = time.time()

    check_generate_params_args() # Unique error message for --generate_params to print species names
    parser = parse_args(__version__)
    known_args, remaining_args = parser.parse_known_args()

    if known_args.generate_params is not None: # if --generate_params
        print(f"Retrieving params from template...")
        generateParams(known_args.generate_params, known_args.dir)
        return
    
    if known_args.download_sample_data: # if --download_sample_data
        success = download_sample_data(force=False, quiet=False, target_dir=known_args.dir)
        sys.exit(0 if success else 1)
    
    if known_args.Bmap is not None: # if --Bmap
        args = parseBmapArgs(remaining_args)
        flat_b, flat_chrom = positionsBstats(args, known_args.Bmap)
        if args.plot_distribution:
            plotChromB(flat_b, flat_chrom, args.plot_distribution, args.quiet)
        return

    print(f"= Calculating relative diversity (B) for all neutral sites across the genome. = = =")

    if known_args.genome: # Run genome Bcalc
        args = parseGenomeArgs(remaining_args)
        os.environ["BCALC_params"] = args.params  # Save params to global
        from Bvalcalc.core.genomeBcalc import genomeBcalc
        genomeBcalc(args)

    elif known_args.region: # Run region Bcalc
        args = parseRegionArgs(remaining_args)
        os.environ["BCALC_params"] = args.params  # Save params to global
        from Bvalcalc.core.regionBcalc import regionBcalc
        output_data, block_ranges, rec_rate_per_chunk_in_region, chunk_size = regionBcalc(args, known_args.region)
        if getattr(args, 'plot', True):
            plotB(b_values_input=output_data, caller="chromosome", output_path=args.plot, quiet=args.quiet, gene_ranges=block_ranges, neutral_only=args.neutral_only, rec_rates=rec_rate_per_chunk_in_region, chunk_size=chunk_size)

    elif known_args.gene: # Run gene Bcalc
        args = parseGeneArgs(remaining_args)
        os.environ["BCALC_params"] = args.params  # Save params to global
        from Bvalcalc.core.geneBcalc import geneBcalc
        output_data = geneBcalc(args) # Capture the output from geneBcalc
        if getattr(args, 'plot', False): # If the --plot flag was provided, call plotB with geneBcalc's output.
            plotB(b_values_input=output_data, caller="gene", output_path=args.plot, quiet=args.quiet)

    elif known_args.site: # Run single site Bcalc
        args = parseSiteArgs(remaining_args)
        os.environ["BCALC_params"] = args.params  # Save params to global
        from Bvalcalc.core.siteBcalc import siteBcalc
        siteBcalc(args)

    print(f"= B value calculated in {time.time() - start_time:.2f} seconds. = = =")

if __name__ == "__main__":
    main()


import csv

def load_chr_sizes(file_path):
    chr_size_dict = {}
    valid_rows = 0

    with open(file_path, newline='') as csvfile:
        reader = csv.reader(csvfile)

        for row in reader:
            if len(row) < 2:
                continue

            chr_name = row[0].strip()
            size_str = row[1].strip()

            try:
                chr_size = int(size_str)
            except ValueError:
                continue

            chr_size_dict[chr_name] = chr_size
            valid_rows += 1

    if valid_rows == 0:
        raise ValueError(f"No valid chromosome size entries found in '{file_path}'. Expecting format: string,int per line.")

    return chr_size_dict



import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.ticker as ticker
import numpy as np
from matplotlib.collections import LineCollection
from matplotlib import gridspec  # for rec rate strip
import sys
from matplotlib.lines import Line2D

def plotB_figures_200kb(b_values_input, caller, output_path, quiet, gene_ranges=None, neutral_only=False, rec_rates=None):
    if not quiet:
        print('====== P L O T T I N G . . . =======================ss')

    B_uncorrected = None
    B_observed = None
    legend_name_blue = "Calculated B (intergenic)"
    legend_name_orange = "Uncorrected B"
    legend_name_dot = "Observed (simulations)"
    legend_name_black = "Calculated B (synonymous)"

    # chr_200kb AKA 200kb Genome
    # poetry run Bvalcalc --region chr_200kb:1-200000 --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --plot /Users/jmarsh96/Desktop/Bcalc/Figures/chr_200kb.png
    B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/200kb_all.pi"
    title_name = 'B for 200 kb genome with 10 selected elements'

    # chr_200kb_recmap AKA 200kb Rec Map
    # poetry run Bvalcalc --region chr_200kb:1-200000 --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --plot /Users/jmarsh96/Desktop/Bcalc/Figures/chr_200kb_recmap.png
    # B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/200kb_recmap_all.pi"
    # title_name = 'B for 200 kb genome with recombination map'
    Genome = True

    if B_uncorrected is not None:
        load_B_uncorrected(B_uncorrected)
    if B_observed is not None:
        load_B_observed(B_observed)

    mpl.rcParams['font.family'] = ['Helvetica', 'DejaVu Sans', 'Arial']
    if 'seaborn-v0_8-whitegrid' in plt.style.available:
        plt.style.use('seaborn-v0_8-whitegrid')
    else:
        print("'seaborn-whitegrid' style not found, using 'ggplot' as fallback.")
        plt.style.use('ggplot')
        mpl.rcParams['axes.facecolor'] = 'white'
        mpl.rcParams['figure.facecolor'] = 'white'
        mpl.rcParams['grid.color'] = 'grey'
        mpl.rcParams['grid.linestyle'] = '--'
        mpl.rcParams['grid.linewidth'] = 0.5
    mpl.rcParams['axes.edgecolor'] = 'black'

    if rec_rates is None:
        fig, ax = plt.subplots(figsize=(10, 6))
    else:
        height_ratios = [10, 0.5]
        fig = plt.figure(figsize=(10, 6))
        gs = gridspec.GridSpec(nrows=2, ncols=1, height_ratios=height_ratios)
        ax = fig.add_subplot(gs[0])

    if caller == "chromosome":
        positions = b_values_input['Position']
        b_vals = b_values_input['B']
        chrom = b_values_input['Chromosome'][0] if 'Chromosome' in b_values_input.dtype.names else 'unknown'

        x = positions
        y = b_vals

        max_points = 10000
        if len(x) > max_points:
            idx = np.linspace(0, len(x) - 1, max_points).astype(int)
            x = x[idx]
            y = y[idx]
        ax.plot(x, y, color='black', lw=1.5, alpha=1, zorder=1, label=legend_name_black)
        ax.plot(x, y, color='blue', lw=1.5, alpha=1, zorder=2, label=legend_name_blue)
        ax.set_xlim(x.min() - 1, x.max())
        ax.set_ylim(0.4, 1.0)

    elif caller == "gene":
        x = b_values_input[:, 0]
        y = b_values_input[:, 1]
        ax.plot(x, y, color='blue', lw=1.5, alpha=0.8, label=legend_name_blue)
        ax.set_xlim(x.min() - 1, x.max())
        ax.set_ylim(0.4, 1.0)

    ax.set_ylabel('Expected diversity relative to neutral evolution (B)', fontsize=13)
    if caller == "chromosome":
        ax.set_title(f'{title_name}', fontsize=15, fontweight='bold')
        if rec_rates is not None:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13, labelpad=40)
        else:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13)
    else:
        ax.set_xlabel('Distance from single selected element of size 10 kb', fontsize=13)
        ax.set_title(title_name, fontsize=15, fontweight='bold')

    if B_uncorrected is not None:
        uncorrected_data = load_B_uncorrected(B_uncorrected)
        if uncorrected_data is not None and len(uncorrected_data) > 0:
            ax.plot(
                uncorrected_data["Distance"],
                uncorrected_data["B"],
                color='orange',
                lw=1.5,
                alpha=0.5,
                label=legend_name_orange
            )

    if Genome is True:
        if B_observed is not None:
            print(f"Add code")
            observed_data = load_B_observed(B_observed)
            observed_data = observed_data[observed_data["B"] > 0.001]

    else:
        if B_observed is not None:
            observed_data = load_B_observed(B_observed)

    if observed_data is not None and len(observed_data) > 0:
        ax.scatter(
            observed_data["Distance"] + 500,
            observed_data["B"],
            color='#f57616',
            s=20,
            alpha=0.9,
            label=legend_name_dot,
            zorder=3
        )

    ax.tick_params(axis='both', which='major', labelsize=10)

    if gene_ranges is not None and len(gene_ranges) > 0:
        ymin, ymax = ax.get_ylim()
        bar_y = ymin - (ymax - ymin) * 0.05
        ax.set_ylim(bar_y, ymax)

        segments = [((int(start), bar_y), (int(end), bar_y)) for _, start, end in gene_ranges]
        ax.add_collection(LineCollection(segments, colors='black', linewidths=30))

        x = np.asarray(x, dtype=float)
        gene_mask = np.zeros_like(x, dtype=bool)
        for _, start, end in gene_ranges:
            start = int(start)
            end = int(end)
            gene_mask |= (x >= start) & (x <= end)
        idx = np.where(gene_mask)[0]
        splits = np.where(np.diff(idx) != 1)[0] + 1
        for seg in np.split(idx, splits):
            if len(seg) > 1:
                coords = np.column_stack((x[seg], y[seg]))
                ax.add_collection(LineCollection([coords], colors='black', linewidths=1.5, label=legend_name_black if seg is splits[0] else "_nolegend_", zorder=2))

    ax.xaxis.set_major_formatter(
        ticker.FuncFormatter(lambda value, pos:
            f"{int(value)} bp" if value < 1e3 else (
                f"{value/1e6:.2f} Mb" if value >= 1e6 else f"{int(value/1e3)} kb"
            )
        )
    )
    from matplotlib.colors import LinearSegmentedColormap
    magenta_map = LinearSegmentedColormap.from_list("custom_magenta", ["white", "#C54B8C"])


    if rec_rates is not None and caller == "chromosome":
        ax_rec = fig.add_subplot(gs[1], sharex=ax)
        ax_rec.set_yticks([])
        ax_rec.tick_params(axis='x', which='major', labelsize=9)
        rec_img = np.expand_dims(rec_rates, axis=0)
        min_pos = positions.min()
        extent = [min_pos, min_pos + len(rec_rates) * 20000, 0, 1]
        ax_rec.imshow(rec_img, aspect='auto', extent=extent, cmap=magenta_map, origin='lower', zorder=2, vmin=0, vmax=np.max(rec_rates))
        ax_rec.set_frame_on(False)
        plt.setp(ax.get_xticklabels(), visible=False)

    if rec_rates is None:
        plt.tight_layout()
    else:
        fig.subplots_adjust(hspace=0.01, bottom=0.12)

    ax.legend(loc="lower right", bbox_to_anchor=(1, 0.04), fontsize=10, frameon=True)

    print("Observed dot at:", observed_data)
    print("Calculated B at 19999 and 20000:", b_values_input[19997:20101])

    plt.savefig(output_path, dpi=300)
    print(f"Plot saved to {output_path}")

import numpy as np

def load_B_uncorrected(file_path):
    try:
        data = np.genfromtxt(file_path, delimiter=",", names=True, dtype=[("Distance", float), ("B", float)])
        return data
    except Exception as e:
        print(f"Error reading B_observed file: {file_path}")
        raise e

def load_B_observed(file_path):
    try:
        raw = np.genfromtxt(file_path, delimiter=",", names=True)
        import pandas as pd
        df = pd.DataFrame(raw)
        grouped = df.groupby("start", as_index=False)["pi"].mean()
        grouped["Distance"] = grouped["start"]
        grouped["pi"] = grouped["pi"] / 0.012
        grouped.rename(columns={"pi": "B"}, inplace=True)
        result = grouped[["Distance", "B"]].to_records(index=False)
        return result
    except Exception as e:
        print(f"Error reading B_uncorrected file: {file_path}")
        raise e



from Bvalcalc.core.chromBcalc import chromBcalc
from Bvalcalc.utils.load_bed_gff import load_bed_gff
from Bvalcalc.utils.load_chr_sizes import load_chr_sizes
from Bvalcalc.utils.load_Bmap import load_Bmap
from Bvalcalc.core.calculateB import calculateB_unlinked
import numpy as np

def genomeBcalc(args):    

    allblockstart, allblockend, allblockchrom = load_bed_gff(args.bedgff) # Read BED/GFF, return start and end of conserved elements

    unique_chromosomes = np.unique(allblockchrom) # Move BED/GFF handler here
    if args.chr_sizes is not None: 
        chr_sizes = load_chr_sizes(args.chr_sizes)  # <-- Path to the sizes CSV file

    import Bvalcalc.utils.dfe_helper as dfe_helper
    dfe_helper.GAMMA_DFE = args.gamma_dfe # Update DFE if --gamma_dfe
    dfe_helper.CONSTANT_DFE = args.constant_dfe # Update DFE if --constant_dfe

    print("Chromosomes loaded:", unique_chromosomes) ## Now, loop over each chromosome and save B output

    if args.prior_Bmap is not None:
        prior_chromosomes, prior_positions, prior_b_values = load_Bmap(file_path = args.prior_Bmap)
        if not args.quiet: print(f"Using prior B values from {args.prior_Bmap}")

    if args.out is not None: # Overwrite existing file with header
        with open(args.out, 'w') as out_f:
            out_f.write("Chromosome,Start,B\n")

    for i in np.arange(0,len(unique_chromosomes)):

        mask = allblockchrom == unique_chromosomes[i]
        blockstart, blockend = allblockstart[mask], allblockend[mask]
        chromosome = unique_chromosomes[i]
        unlinked_blockstart, unlinked_blockend = allblockstart[~mask], allblockend[~mask]
        unlinked_L = np.sum(unlinked_blockend-unlinked_blockstart)
        unlinked_B = calculateB_unlinked(unlinked_L)

        if args.prior_Bmap is not None: 
            prior_mask = (prior_chromosomes == chromosome)
            prior_pos = prior_positions[prior_mask]
            prior_b = prior_b_values[prior_mask]
        else:
            prior_pos, prior_b = None, None

        if args.chr_sizes is not None: chr_size = chr_sizes.get(chromosome)
        else: chr_size = None
        chromBcalc(args, blockstart, blockend, chromosome, unlinked_B, prior_pos, prior_b, calc_start=None, calc_end=None, chr_size=chr_size, caller="genomeBcalc")

    return


import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
import numpy as np

def plotChromB(flat_b, flat_chrom, output_path, quiet):
     if not quiet: print('====== P L O T T I N G . . . =======================')

     # Set overall style
     sns.set_theme(style="whitegrid", font='Helvetica')
     mpl.rcParams['axes.edgecolor'] = 'black'

     fig, ax = plt.subplots(figsize=(12, 6))

     # Unique chromosomes plus an "All" category
     chroms = list(np.unique(flat_chrom))
     chroms.append('All')

     # Extend arrays to include the combined "All" data
     all_label = np.array(['All'] * len(flat_b), dtype='<U20')
     ext_chrom = np.concatenate([flat_chrom, all_label])
     ext_b = np.concatenate([flat_b, flat_b])

     # Boxenplot including the "All" column, with hue to align palette
     # Use k_depth='tukey' to speed up computation by using Tukey's five-number summary
     sns.boxenplot(
         x=ext_chrom,
         y=ext_b,
         order=chroms,
         hue=ext_chrom,
         palette='pastel',
         showfliers=False,
         legend=False,
         k_depth='tukey',
         ax=ax
     )

     # Final plot adjustments
     ax.set_xlabel('Chromosome', fontsize=13)
     ax.set_ylabel('Expected diversity relative to neutral evolution (B)', fontsize=13)
     ax.set_title('Distribution of B by Chromosome', fontsize=15, fontweight='bold')
     ax.tick_params(axis='x', rotation=45, labelsize=10)
     ax.tick_params(axis='y', labelsize=10)

     plt.tight_layout()
     plt.savefig(output_path, dpi=300)
     print(f"Plot saved to {output_path}")



## Population genetic parameters for the simulated or empirical population
## Accurate estimation requires accurate and appropriate parameters
##
## e.g. Bvalcalc --params path/to/ExampleParams.py
##
## Core parameters
f = 0.97 # Selfing rate (f = S/(2-S); Wright's inbreeding coefficient) [1]
x = 1 # Scaling factor (N,u,r), keep as 1 unless calculating for rescaled simulations
Nanc = 1e5 / (1+f) / x # Ancestral population size [2]
r = 7.465e-8 * (1-f) * x # Recombination (crossover) rate per bp, per generation (sex-averaged) [3]
u = 6.95e-9 * x # Mutation rate (all types) per bp, per generation [4]
g = r * 50 * (1-f) * x # Gene conversion initiation rate per bp, per generation [5]
k = 553 # Gene conversion tract length (bp) [5]
## DFE parameters for ALL sites in annotated regions (Sum must equal 1)
f0 = 0.2 # Proportion of effectively neutral mutations with 0 <= |2Ns| < 1 (Note that 2Ns<5 does not contribute to BGS) [6]
f1 = 0.14 # Proportion of weakly deleterious mutations with 1 <= |2Ns| < 10 [6]
f2 = 0.44 # Proportion of moderately deleterious mutations with 10 <= |2Ns| < 100 [6]
f3 = 0.22 # Proportion of strongly deleterious mutations with |2Ns| >= 100 [6]
## Demography parameters
Ncur = 0.5 * Nanc # Current population size (!Requires --pop_change) [2]
time_of_change = 1 # Time in Nanc generations ago that effective population size went from Nanc to Ncur (!Requires --pop_change) [2]
## Advanced DFE parameters 
h = 0.5 + (f-0.5*f) # Dominance coefficient of selected alleles [Naive value]
mean, shape, proportion_synonymous = 500, 0.5, 0.3 # Gamma distribution of DFE to discretize and replace f0-f3 [mean (2Ns), shape, proportion synonymous] (!Requires --gamma_dfe) [Naive value]## Literature cited
# [1] Platt et al. 2010 doi: 10.1371/journal.pgen.1000843
# [2] Durvasula et al. 2017 doi: 10.1073/pnas.1616736114
# [3] Rowan et al. 2019 doi: 10.1534/genetics.119.302406
# [4] Weng et al. 2018 doi: 10.1534/genetics.118.301721
# [5] Yang et al. 2012 doi: 10.1073/pnas.1211827110
# [6] Hämälä and Tiffin 2020 doi: 10.1534/genetics.120.303335


import numpy as np
from Bvalcalc.utils.dfe_helper import get_DFE_params
from scipy.optimize import root_scalar
from scipy.integrate import trapezoid

_params_cache: dict | None = None
_cache_args: tuple[str | None, bool, bool] | None = None

def get_params(
    params_path: str | None = None,
    gamma_dfe: bool = False,
    constant_dfe: bool = False,
):
    """
    Loads DFE parameters from the provided population genetic parameters file.
    Caches on (params_path, gamma_dfe, constant_dfe) and rebuilds whenever
    any of those three inputs change.
    """
    global _params_cache, _cache_args
    from Bvalcalc.utils.dfe_helper import GAMMA_DFE, CONSTANT_DFE
    
    # Get the actual params path that will be used
    actual_params_path = params_path
    if actual_params_path is None:
        import os
        actual_params_path = os.environ.get("BCALC_params")
    
    # Include global DFE state in cache key
    key = (actual_params_path, gamma_dfe or GAMMA_DFE, constant_dfe or CONSTANT_DFE)
    if _cache_args != key:
        _params_cache = get_DFE_params(params_path, gamma_dfe, constant_dfe)
        _cache_args = key
    return _params_cache

def calculateB_linear(distance_to_element: int, length_of_element: int, params: dict | None = None):
    """
    Calculate B due to purifying selection acting on a linked selected element of arbitrary length, assuming a constant crossover and gene conversion rate (analytical solution).
 
    Parameters 
    ----------
    distance_to_element: int
        Distance (bp) from the neutral site to the nearest edge of the selected element.
    length_of_element: int
        Length (bp) of the selected element.
    params : dict
        Required parameters from ``get_params()``, only kept as default (None) when being called by CLI,
        in which case parameters are sourced from the params file directly.
    """    
    with np.errstate(divide='ignore', invalid='ignore'):
        if params is None:
            params = get_params()
        r, u, g, k, t1, t1half, t2, t3, t4, f1, f2, f3, f0, t_constant = params["r"], params["u"], params["g"], params["k"], params["t1"], params["t1half"], params["t2"], params["t3"], params["t4"], params["f1"], params["f2"], params["f3"], params["f0"], params["t_constant"]

        C = (1.0 - np.exp(-2.0 * r * distance_to_element)) / 2.0 # cM
        U = length_of_element * u
        if g == 0:
            a = C # RECOMBINATION IN Y
            b = C + (r * length_of_element) # RECOMBINATION IN X
        elif g > 0:
            a, b = get_a_b_with_GC(C, distance_to_element, length_of_element)

        if t_constant: #If --constant_dfe is active
            E_constant = calculate_exponent(t_constant, t_constant, U, a, b)
            B = np.exp(-1.0 * E_constant)
            return np.where(length_of_element == 0, 1.0, B)
        
        E_f1 = calculate_exponent(t1half, t2, U, a, b)
        E_f2 = calculate_exponent(t2, t3, U, a, b)
        E_f3 = calculate_exponent(t3, t4, U, a, b)

        E_bar = ( # Sum over the DFE
            f0 * 0.0
            + f1 * ((t1half - t1) / (t2 - t1)) * 0.0
            + f1 * ((t2 - t1half) / (t2 - t1)) * E_f1
            + f2 * E_f2
            + f3 * E_f3)

        B = np.exp(-1.0 * E_bar)
        
    return np.where(length_of_element == 0, 1.0, B)

def calculateB_recmap(distance_to_element, length_of_element, 
                      rec_distances = None, rec_lengths = None, 
                      gc_distances = None, gc_lengths = None, params = None):
    """
    Calculate the B value WITH REC MAP for a single functional element at the focal site,
    summing over the DFE while consolidating the intermediate calculations.
    """    
    with np.errstate(divide='ignore', invalid='ignore'):
        if params is None:
            params = get_params()
        r, u, g, k, t1, t1half, t2, t3, t4, f1, f2, f3, f0, t_constant = params["r"], params["u"], params["g"], params["k"], params["t1"], params["t1half"], params["t2"], params["t3"], params["t4"], params["f1"], params["f2"], params["f3"], params["f0"], params["t_constant"]
        # rec_distances is the length of the element * rec rate in each spanned region. 
        
        if rec_distances is not None:
            rec_adjusted_length_of_element = rec_lengths 
            rec_adjusted_distance_to_element = rec_distances
        else:
            rec_adjusted_length_of_element = length_of_element
            rec_adjusted_distance_to_element = distance_to_element
        
        if gc_distances is not None:
            local_g = (gc_lengths + gc_distances)/(length_of_element + distance_to_element) * g
        else:
            local_g = g
        
        C = (1.0 - np.exp(-2.0 * r * rec_adjusted_distance_to_element)) / 2.0 # cM
        U = length_of_element * u
        if g == 0:
            a = C
            b = C + r * rec_adjusted_length_of_element # cM
        elif g > 0:
             a, b = get_a_b_with_GC_andMaps(C, y=distance_to_element, l=length_of_element, 
                                            rec_l=rec_adjusted_length_of_element, local_g = local_g)

        if t_constant: #If --constant_dfe is active
            E_constant = calculate_exponent(t_constant, t_constant, U, a, b)
            B = np.exp(-1.0 * E_constant)
            return np.where(length_of_element == 0, 1.0, B)
        
        E_f1 = calculate_exponent(t1half, t2, U, a, b)
        E_f2 = calculate_exponent(t2, t3, U, a, b)
        E_f3 = calculate_exponent(t3, t4, U, a, b)

        E_bar = ( # Sum over the DFE
            f0 * 0.0
            + f1 * ((t1half - t1) / (t2 - t1)) * 0.0
            + f1 * ((t2 - t1half) / (t2 - t1)) * E_f1
            + f2 * E_f2
            + f3 * E_f3)    

        B = np.exp(-1.0 * E_bar)
        
    return np.where(length_of_element == 0, 1.0, B)

def calculateB_unlinked(unlinked_L: int, params: dict | None = None):
    """
    Calculate B due to purifying selection at unlinked sites (numerical integration over DFE).

    Parameters
    ----------
    unlinked_L : float
        Cumulative count of selected sites in unlinked regions.
    params : dict
        Required parameters from ``get_params()``, only kept as default (None) when being called by CLI,
        in which case parameters are sourced from the params file directly.
    """
    if params is None:
        params = get_params()

    u, t1, t1half, t2, t3, t4, f0, f1, f2, f3, t_constant = params["u"], params["t1"], params["t1half"], params["t2"], params["t3"], params["t4"], params["f0"], params["f1"], params["f2"], params["f3"], params["t_constant"]
    
    if t_constant: #If --constant_dfe is active    

        unlinked_B  = np.exp(-8 * u * 1.0 * unlinked_L * (t_constant/(1 + t_constant)**2))
        return unlinked_B    

    f1_above_cutoff = f1 * ((t1half - t1) / (t2 - t1))

    sum_f1 = (f1_above_cutoff / (t2 - t1half)) * (np.log((1 + t2) /(1 + t1half)) + (1 / (1 + t2)) - (1 / (1 + t1half)))
    sum_f2 = (f2 / (t3 - t2)) * (np.log((1 + t3) /(1 + t2)) + (1 / (1 + t3)) - (1 / (1 + t2)))
    sum_f3 = (f3 / (t4 - t3)) * (np.log((1 + t4) /(1 + t3)) + (1 / (1 + t4)) - (1 / (1 + t3)))
    
    unlinked_B  = np.exp(-8 * u * 1.0 * unlinked_L * (sum_f1 + sum_f2 + sum_f3))

    return unlinked_B


##



## Helper functions

def calculate_exponent(t_start, t_end, U, a, b):
    """"
    Helper to calculate the exponent using "a" and "b"
    """
    a, b, U = np.asarray(a), np.asarray(b), np.asarray(U)

    if U.size == 0: return 0 # If e.g. f1 proportion is 0, no need to calculate exponent
    
    if t_end == t_start: # If --constant_dfe
        E = (U / (a - b)) * (
            a / (a + (1 - a) * t_start) -
            b / (b + (1 - b) * t_start)
        )
    else: # Using discretized DFE (f0,f1,f2,f3 or --gamma_dfe)
        E1 = ((U * a) 
                / ((1 - a) * (a - b) * (t_end - t_start))) * np.log((a + (t_end * (1 - a))) 
                / (a + (t_start * (1 - a))))
        E2 = -1.0 * ((U * b) 
                / ((1 - b) * (a - b) * (t_end - t_start))) * np.log((b + ((1 - b) * t_end)) 
                / (b + ((1 - b) * t_start)))
    
        E = np.asarray(E1 + E2)

    rec_0_mask = np.isclose(a, b)  # Get mask for where recombination rate = 0 within the gene
    if rec_0_mask.any(): # 4a) If a_arr is scalar (0‐d), compute limit once as scalar
        if a.ndim == 0:
            limit_factor = (1 / ((t_end - t_start)*(1-a)**2)) * ( # Calculate exponent with 0 recombination between gene and site, avoiding limits
                np.log((a + (1 - a) * t_end) 
                       / (a + (1 - a) * t_start))
                + a / (a + (1 - a) * t_end)
                - a / (a + (1 - a) * t_start))
            if t_start == t_end: limit_factor = t_start / (a + (1 - a) * t_start)**2 # If --constant_dfe
            # Broadcast scalar limit_factor to all masked positions
            E[rec_0_mask] = U[rec_0_mask] * limit_factor  # Get corresponding U for the numerator and plug back into E array to replace nan's

        else: # 4b) If a_arr is array, compute limit for each masked element
            ae = a[rec_0_mask]  # array of a_i where a_i ≈ b_i
            limit_factor = (1 / ((t_end - t_start)*(1-ae)**2)) * ( # Calculate exponent with 0 recombination between gene and site, avoiding limits
                np.log((ae + (1 - ae) * t_end) 
                       / (ae + (1 - ae) * t_start))
                + ae / (ae + (1 - ae) * t_end)
                - ae / (ae + (1 - ae) * t_start))
            if t_start == t_end: limit_factor = t_start / (ae + (1 - ae) * t_start)**2 # If --constant_dfe
            ## REPLACED BELOW WITH THE NEW LINE TO FIX FAR GENE ISSUE, MAY NEED TO REVERT
            E[rec_0_mask] = U[rec_0_mask] * limit_factor
            # Match array of limit_factor to corresponding positions in E (where rec_0_mask has True);'l;'l''
            # if len(rec_0_mask[False]) == 0:
            #     # print(f"Need to fix --gene when r = 0, see calculateB ~line 176") Fixed??
            #     E[rec_0_mask] = U * limit_factor
            # else:
            #     E[rec_0_mask] = U[rec_0_mask] * limit_factor  # Get corresponding U for the numerator and plug back into E array to replace nan's

    return E

def get_a_b_with_GC(C, y, l):
        with np.errstate(divide='ignore', invalid='ignore'):
            params = get_params()
            r, u, g, k, t1, t1half, t2, t3, t4, f1, f2, f3, f0 = params["r"], params["u"], params["g"], params["k"], params["t1"], params["t1half"], params["t2"], params["t3"], params["t4"], params["f1"], params["f2"], params["f3"], params["f0"]
            proportion_nogc_a = np.where(k < y + l, # When GC includes neutral site, this is proportion of the gene it includes
                                        np.maximum((0.5*(k-y)/l), 0),
                                        1-(y + l)/(2 * k)
                                        )

            proportion_nogc_b = np.where(k < y + l, # When GC includes gene site, this is probability the tract includes neutral site of interest 
                                    1/(2*k) * np.maximum(k-y+1,0) * np.maximum(k - y, 0) / l, # When overshooting not possible
                                    (k - y - 0.5 * l) / k) # When overshooting possible
            
        
        a = np.where(k < y, 
            C + (2 * g * k), # Probability of GC on neutral site, where overlap with element not possible
            C + (2 * g * (y) + # When overlap possible this is probability gc is in neutral but doesn't include any of element
                g * (k - y) * # Probability gc is in neutral and includes some element (remaining probability from above)
                (1 - proportion_nogc_a) # Proportion of gene that gc breaks linkage with when it includes some element
        ))
        b = C + (r * l) + (2 * g * k) * (1 - (1-proportion_nogc_a)*proportion_nogc_b) #* prop k out

        return a, b

def get_a_b_with_GC_andMaps(C, y, l, rec_l, local_g):
        params = get_params()
        r, u, g, k, t1, t1half, t2, t3, t4, f1, f2, f3, f0 = params["r"], params["u"], params["g"], params["k"], params["t1"], params["t1half"], params["t2"], params["t3"], params["t4"], params["f1"], params["f2"], params["f3"], params["f0"]
        with np.errstate(divide='ignore', invalid='ignore'):
            proportion_nogc_a = np.where(k < y + l, # When GC includes neutral site, this is proportion of the gene it includes
                                        np.maximum((0.5*(k-y)/l), 0),
                                        ((y) * (2 * k - (y + l)))/(2 * k * y)
                                        )

            proportion_nogc_b = np.where(k < y + l, # When GC includes gene site, this is probability the tract includes neutral site of interest 
                                    1/(2*k) * np.maximum(k-y+1,0) * np.maximum(k - y, 0) / l,
                                    (k - y - 0.5 * l) / k)
        
        a = np.where(k < y, 
            C + (2 * local_g * k), # Probability of GC on neutral site, where overlap with element not possible
            C + (2 * local_g * (y) + # When overlap possible this is probability gc is in neutral but doesn't include any of element
                local_g * (k - y) * # Probability gc is in neutral and includes some element (remaining probability from above)
                (1 - proportion_nogc_a) # Proportion of gene that gc breaks linkage with when it includes some element
        ))
        b = C + (r * rec_l) + (2 * local_g * k) * (1 - (1-proportion_nogc_a)*proportion_nogc_b) #* prop k out

        return a, b

def calculateB_hri(distant_B, interfering_L, params: dict | None = None):
    """
    Fully vectorized calculation of B' under Hill-Robertson interference.
    """
    if params is None:
        params = get_params()

    Nanc, u, f1, f2 = params["Nanc"], params["u"], params["f1"], params["f2"]

    distant_B = np.atleast_1d(distant_B).astype(float)
    interfering_L = np.atleast_1d(interfering_L).astype(float)

    scalar_input = distant_B.shape == () or distant_B.shape == (1,)

    N0 = distant_B * Nanc
    h = 0.5
    u = 2 * u
    u1 = f1 * u
    u2 = f2 * u
    u_total = u1 + u2

    E_X2_f1 = (1**2 + 1*10 + 10**2) / 3
    E_X2_f2 = (10**2 + 10*100 + 100**2) / 3

    t_sq1 = (h**2 * E_X2_f1) / (4 * N0**2)
    t_sq2 = (h**2 * E_X2_f2) / (4 * N0**2)
    t = np.sqrt((u1 * t_sq1 + u2 * t_sq2) / u_total)

    gamma = 2 * N0 * t
    U = u_total * interfering_L
    alpha2 = 2 * N0 * U
    kappa = 1.0

    def eq4(B, U, gamma, t):
        exp_term = np.exp(-gamma * B)
        num = 0.5 * U * (1 - exp_term)**3
        denom = t * (1 + kappa * exp_term)**3
        return -np.log(B) - num / denom

    def solve_eq4_batched(U, gamma, t, n=500):
        Bgrid = np.linspace(1e-10, 1.0, n)[None, :]
        U = np.asarray(U).reshape(-1, 1)
        gamma = np.asarray(gamma).reshape(-1, 1)
        t = np.asarray(t).reshape(-1, 1)

        fvals = eq4(Bgrid, U, gamma, t)
        signs = np.sign(fvals)
        crossing = np.diff(signs, axis=1) < 0
        idx = np.argmax(crossing, axis=1)

        B_left = Bgrid[0, idx]
        B_right = Bgrid[0, idx + 1]
        f_left = fvals[np.arange(len(U)), idx]
        f_right = fvals[np.arange(len(U)), idx + 1]

        B_root = B_left - f_left * (B_right - B_left) / (f_right - f_left)
        return B_root

    Bval = solve_eq4_batched(U, gamma, t)

    def eq5_vectorized(B, alpha2, gamma, Tmax=100.0, n_steps=2000):
        x = np.linspace(0, Tmax, n_steps)[None, :]  # shape (1, n_steps)
        dx = x[0, 1] - x[0, 0]

        B = B[:, None]
        alpha2 = alpha2[:, None]
        gamma = gamma[:, None]

        f1 = 1 - np.exp(-gamma * B)
        f2 = 1 + kappa * np.exp(-gamma * B)
        A = f1 / f2
        c = 0.5 * alpha2 / gamma * A**3
        d = 2 * gamma * B * (f2 / f1)

        x_broadcasted = np.broadcast_to(x, (B.shape[0], x.shape[1]))
        gx = np.exp(c * (1 - np.exp(-d * x_broadcasted))**2)
        cumI = np.cumsum((gx[:, :-1] + gx[:, 1:]) * 0.5 * dx, axis=1)
        cumI = np.hstack([np.zeros((gx.shape[0], 1)), cumI])

        hx = np.exp(-B * cumI)
        Bprime = B[:, 0] * trapezoid(hx, x[0], axis=1)
        return Bprime

    Bprime = eq5_vectorized(Bval, alpha2, gamma)
    return Bprime[0] if scalar_input else Bprime
